#ifndef RIGID_TRANSFORM
#define RIGID_TRANSFORM
#include<Eigen/Core>
#include<Eigen/Geometry>
using namespace Eigen;
#include<iostream>
using namespace Eigen;

template<typename _Scalar>
class RigidTransform
{
  public:
  /** the scalar type of the coefficients */
  typedef _Scalar Scalar;

  /** the type of the Coefficients 4-vector */
  typedef Matrix<Scalar, 4, 1> Coefficients;
  /** the type of a 3D vector */
  typedef Matrix<Scalar,3,1> VectorType;
  /** the equivalent rotation matrix type */
  typedef Matrix<Scalar,3,3> MatrixType;
  /** the equivalent angle-axis type */
  typedef AngleAxis<Scalar> AngleAxisType;
  /** the equivalent quaternion type */
  typedef Quaternion<Scalar> QuatType;
  /** the equivalent translation type */
  typedef Translation<Scalar, 3> TranslationType;  
  /** the equivalent line type */
  typedef ParametrizedLine<Scalar,3> LineType;
  /** the equivalent real data type */
  typedef typename NumTraits<Scalar>::Real RealScalar;
  
  protected:
    QuatType quats[2];
   
public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW_IF_VECTORIZABLE_FIXED_SIZE(_Scalar,8)
  
  
 inline RigidTransform()
	{
	this->quats[0].setIdentity();
    this->quats[1].coeffs().setZero();
	}

 inline RigidTransform(const RigidTransform &other)
	{ 
	quats[0]=other.quats[0];
  	quats[1]=other.quats[1];
	}
 inline RigidTransform(const QuatType &prim, const QuatType &dual)
	{ 
	quats[0]=prim;
  	quats[1]=dual;
	}
  inline RigidTransform & operator=(const RigidTransform &other)
	  {
	  quats[0]=other.quats[0];
	  quats[1]=other.quats[1];
	  return *this;
	  }
  inline RigidTransform(const AngleAxisType& aa)
	  { 
	  VectorType axis_=aa.axis();
	  axis_.normalize();
	  quats[0]=AngleAxisType(aa.angle(), axis_);
//	  std::cout<<"\ndebugging angleaxis cons() "<<quats[0].coeffs();
	  quats[1].coeffs().setZero();
	  }
  inline RigidTransform(const TranslationType& delta)
	  {
	  quats[0].setIdentity();
	  quats[1].coeffs().template start<3>()=delta.vector()*(0.5f);
	  quats[1].w()=1.0f;
	  }
  inline RigidTransform(Scalar angle, const LineType& line)
	  {
	  VectorType dirn=line.direction();
	  dirn=dirn.normalize();
	  quats[0]=AngleAxisType(angle, dirn);
	  VectorType temp=sin(angle*0.5f) * line.through().cross(dirn);
	  quats[1]=QuatType(0.0f, temp.x(), temp.y(), temp.z());
	  }        
  inline RigidTransform operator+(const RigidTransform& other) const
	  {
	  RigidTransform temp;
	  temp.quats[0].coeffs()=quats[0].coeffs() + other.quats[0].coeffs();
	  temp.quats[1].coeffs()=quats[1].coeffs() + other.quats[1].coeffs();
	  return temp;
	  }	  
  inline RigidTransform operator*(const RigidTransform& other) const
	  {
	  RigidTransform temp;
	  temp.quats[0]=(quats[0] * other.quats[0]);
	  temp.quats[1].coeffs()=(quats[1] * other.quats[0]).coeffs() + (quats[0] * other.quats[1]).coeffs();
	  return temp;
	  }
  inline RigidTransform operator*(const AngleAxisType& aa) const
	  {
	  RigidTransform temp(aa);
	  return (*this)*temp;	  
	  }
  inline RigidTransform operator*(const TranslationType& delta) const
	  {
	  RigidTransform temp(delta);
	  return (*this)*temp;	  
	  }	  
  inline RigidTransform operator*(Scalar x) const
	  {
	  RigidTransform temp;
	  temp.quats[0].coeffs()=quats[0].coeffs() * x;
	  temp.quats[1].coeffs()=quats[1].coeffs() * x;
	  return temp;    
	  }
  MatrixType rotation() const
  	{
	  return quats[0].toRotationMatrix();
	  }
  VectorType translation() const
	  {
	  VectorType temp=(quats[1] * (quats[0].conjugate())).vec();
	  return temp*2.0f;
	  }
  void normalize()
	  {
	  RealScalar normalization_prim=quats[0].norm();
	  RealScalar inv_prim=1.0/normalization_prim;
	  RealScalar normalization_dual=quats[0].coeffs().dot(quats[1].coeffs())*inv_prim;
	  RealScalar inv_dual=-(normalization_dual)*inv_prim*inv_prim;
	  quats[0].coeffs()=inv_prim*quats[0].coeffs();
	  quats[1].coeffs()=(quats[0].coeffs()*inv_dual)+(quats[1].coeffs()*inv_prim);
	  } 
  inline static RigidTransform Identity() 
    { 
    RigidTransform temp;
    temp.quats[0].setIdentity();
    temp.quats[1].coeffs().setZero();
    return temp;
    }
  inline RigidTransform& setIdentity() 
    {
    this->quats[0].setIdentity();
    this->quats[1].coeffs().setZero();
    return *this; 
    }
	     
friend inline RigidTransform operator*(const AngleAxisType& aa, const RigidTransform& other)
  {
  RigidTransform temp(aa);
  return temp*other;
  }	  	 
friend inline RigidTransform operator*(const Scalar& x, const RigidTransform& other)
  {
  return other*x;
  }	  	 
friend inline RigidTransform operator*(const TranslationType& delta, const RigidTransform& other)
  {
  RigidTransform temp(delta);
  return temp*other;
  }	  

};
  
#endif
