#include <Eigen/Core>
#include <Eigen/Geometry>
#include <string>
#include <sstream>

#include <renderer.h>
#include <drawScene.h>
#include <tonemap.h>
#include <RigidTransform.h>
#include <grating.h>
#include <diffraction.h>

//#define GRATING_TYPE_BINARY
#define GRATING_TYPE_APERTURE
//#define GRATING_TYPE_APERTURE_GRATING
//#define GRATING_TYPE_LENS
//#define GRATING_TYPE_YOUNG
//#define GRATING_TYPE_YOUNG_APERTURE

//#define CAM_TYPE_ZOOMIN
//#define CAM_TYPE_ZOOMIN_LOTS
//#define CAM_TYPE_ZOOMIN_LOTS_MORE
//#define CAM_TYPE_ZOOMOUT
//#define CAM_TYPE_ZOOMOUT_LOTS
//#define CAM_TYPE_QUADRANT
#define CAM_TYPE_SEE_GRATING

using namespace std;
using namespace Eigen;

//essential definitions for our object manipulation
typedef RigidTransform<float> RigidTransformF;  	       
typedef Translation<float,3> Xlate;
typedef AngleAxis<float> Rotate;		

//Function Declarations:
int initMain();//init sdl, glew,camera,fbo
void keyInput(SDL_Event sdlEv);//keyboard input handler

void render(Texture* tex, ProgramGLSL* shader);//render the frame

void setupSceneCamera(Camera* cam);
void setupFinalCamera(Camera* cam);

//Globals:
Camera cam1;
Uint8 quit;//set this to 1 to quit program

////////MAIN PROGRAM/////////

int main(int argc, char **argv)
    {  
	SDL_Event sdlEv;//to handle events using sdl message passing
	
//read arguments:
	float arg_exposure = 0.0f;
	if (argc > 1)
		{//First argument is the exposure level
		arg_exposure = atof(argv[1]);
		}

//Initialize everything
    int initStatus=initMain(); 
    if (initStatus!=0)
		{
		cout<<"Error: " <<initStatus<<endl;
		return initStatus;
		}
    
//Our renderer object, contains all shaders, vbos, fbos, textures...
    Renderer renderObject;

//Load the shaders
    renderObject.Shaders[WLF_SHADER]		=new ProgramGLSL("./vsh/wlf.vsh","./fsh/wlf.fsh");
    renderObject.Shaders[ACCUMULATE_SHADER]	=new ProgramGLSL("./vsh/quad.vsh","./fsh/accumulate.fsh");
    renderObject.Shaders[FINAL_DRAW_SHADER]	=new ProgramGLSL("./vsh/quad.vsh","./fsh/final.fsh");
    renderObject.Shaders[CLEAR_SHADER]		=new ProgramGLSL("./vsh/quad.vsh","./fsh/clear.fsh");
    renderObject.Shaders[DIFFUSE_SHADER]	=new ProgramGLSL("./vsh/diffuse.vsh","./fsh/diffuse.fsh");

#ifdef GRATING_TYPE_LENS
    Shader * gratingShader = new FragmentShader("./fsh/wlf_lens.fsh");
#elif defined GRATING_TYPE_BINARY
    Shader * gratingShader = new FragmentShader("./fsh/wlf_binarygrating2.fsh");
#elif defined GRATING_TYPE_APERTURE
    Shader * gratingShader = new FragmentShader("./fsh/wlf_aperture.fsh");	
#elif defined GRATING_TYPE_APERTURE_GRATING
    Shader * gratingShader = new FragmentShader("./fsh/wlf_aperturegrating.fsh");
#elif defined GRATING_TYPE_YOUNG
	Shader * gratingShader = new FragmentShader("./fsh/wlf_young.fsh");
#elif defined GRATING_TYPE_YOUNG_APERTURE	
	Shader * gratingShader = new FragmentShader("./fsh/wlf_young_aperture.fsh");
#endif
	renderObject.Shaders[WLF_SHADER]->addShader(gratingShader);	

//    Load the tonemapping shader
    renderObject.Shaders[TONE_MAP_SHADER]=new ProgramGLSL("./vsh/quad.vsh","./fsh/tonemap.fsh");

//check if shaders compiled        
    if(renderObject.validateShaders())
		{
		cerr<<"One of the shaders failed to compile/link!\n";
		cerr<<"Terminating\n";
		return 1;
		}
    
//setup the shader uniforms
    initializeShaderUniforms(renderObject);

//init the tone mapper    
	setTonemapParams(renderObject.Shaders[TONE_MAP_SHADER], 0.0f, arg_exposure, 5.0f,0.0f);

//Initialize the Grating Object to handle grating and lightsource position:
	grating grating1;
	
	grating1.setGratingDimensions(	1.0f,	//grating size (mm)
									16,		//bucket size
									100		//number of samples along x
								);

	grating1.bindToShader(renderObject.Shaders[WLF_SHADER],
						WLF_GRATING_SIZE_UNIFORM,
						WLF_GRATING_POSITION_UNIFORM,
						WLF_GRATING_XAXIS_UNIFORM,
						WLF_GRATING_YAXIS_UNIFORM,
						WLF_BUCKET_SIZE_UNIFORM,
						WLF_SAMPLE_STEP_UNIFORM,
						WLF_LIGHTSOURCE_UNIFORM,
						WLF_LIGHTCOLOR_UNIFORM);

//	grating1.lookat(Vector3f(0.0f,0.0f,500.0f), //look at
//					Vector3f(0.0f,0.0f,-1000.0f),//center X mm away
//					Vector3f(0.0f,1.0f,0.0f));  //up;

	float Grating_Intensity=9000.0f;
//	grating1.setLightColor(Grating_Intensity,Grating_Intensity,Grating_Intensity);
//	grating1.setLightPosn(0.0f,0.0f,-50000.0f);//light at 50 meters away

//set Grating parameters
#ifdef GRATING_TYPE_BINARY									
	//pass the pitch and duty cycle
	renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, 0.1f,0.5f,0.0f,0.0f);	
#elif defined GRATING_TYPE_APERTURE
	//pass the aperture width
	renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, 1.0f,0.0f,0.0f,0.0f);		
#elif defined GRATING_TYPE_APERTURE_GRATING
	//pass the pitch, duty cycle and aperture width
	renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, 0.1f,0.5f,0.5f,0.0f);
#elif defined GRATING_TYPE_LENS
	//pass the focal length and aperture size
	renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, 50.0f,10.0f,0.0f,0.0f);		
#elif defined GRATING_TYPE_YOUNG
	//pass the 2 slit positions
	renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, -0.01,0.01,0.0f,0.0f);			
#elif defined GRATING_TYPE_YOUNG_APERTURE	
	//pass slit_separation/2 and aperture size
	renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, 0.1,0.02,0.0f,0.0f);			
#endif

//Create the scene:
	renderObject.VertexBuffers[SCENE_VBO] = new VertexBuffer;
	createScene(renderObject.VertexBuffers[SCENE_VBO]);

//Create a rendertarget onto which we can draw the screen
    renderObject.RenderTargets[OUTPUT_BUFFER]=new RenderTarget(SCREEN_SIZE, SCREEN_SIZE,false,GL_RGBA32F_ARB);
    
//Create 2 accumulators to add up the results of multiple passes
    renderObject.RenderTargets[ACCUMULATOR1]=new RenderTarget(SCREEN_SIZE, SCREEN_SIZE,false,GL_RGBA32F_ARB);
    renderObject.RenderTargets[ACCUMULATOR2]=new RenderTarget(SCREEN_SIZE, SCREEN_SIZE,false,GL_RGBA32F_ARB);
       
//Setup a scene camera (This is different from the camera we use to navigate)
    setupSceneCamera(&cam1);
    
//setup random number texture:
	renderObject.Textures[RANDOM_TEXTURE] = 0;//create a texture pointer
	setupRandomTexture(renderObject.Textures[RANDOM_TEXTURE], RANDOM_TEXTURE_SIZE, RANDOM_TEXTURE_SIZE);

//load input light filter as a texture:
	renderObject.Textures[FILTER_TEXTURE] = 0;
	loadTextureFromFile(renderObject.Textures[FILTER_TEXTURE], "filter.png");	

//Prepare filenames for multi frame rendering
	string fname_base = "test", fname;
//final result will be stored here
	Texture * finalRender;
	
//	RigidTransformF rtFinal=Xlate(0,0,-20);
//	rtFinal=rtFinal*Rotate(degreesToRadians(20),

//	const int start_frames=10, end_frames=94, frame_skip=1;
	const int start_frames=100, end_frames=200, frame_skip=10;
	
//	grating1.lookat(Vector3f(0.0f,0.0f,0.0f), //look at
//					Vector3f(-100.0f,-100.0f,-100.0f),//center X mm away
//					Vector3f(0.0f,1.0f,0.0f));  //up;

	Grating_Intensity=150.0f;
	grating1.setLightColor(Grating_Intensity,Grating_Intensity,Grating_Intensity);
	
//	float t;
	for (int i=start_frames;i<end_frames;i+=frame_skip)
		{
		// Convert integer to string, and construct the filename
		std::stringstream filenum;
		filenum << i;
		fname = fname_base + filenum.str() + ".png";
		
		int gratePosVary=-100-i;
		grating1.lookat(Vector3f(0.0f,0.0f,0.0f), //look at
				Vector3f(gratePosVary,gratePosVary,gratePosVary),//center X mm away
				Vector3f(0.0f,1.0f,0.0f));  //up;
		
		int lightPosVary=50;//i-50;
		grating1.setLightPosn(-50000.0f-lightPosVary*10,-50000.0f-lightPosVary*lightPosVary,-50000.0f);//light at 50 meters away

//		t=float(i-start_frames)/(end_frames-start_frames);

//	grating1.lookat(Vector3f(0.0f,0.0f,500.0f), //look at
//					Vector3f(0.0f,0.0f,-i),//center X mm away
//					Vector3f(0.0f,1.0f,0.0f));  //up;

//		renderObject.Shaders[WLF_SHADER]->assignUniform3f(WLF_LIGHTSOURCE_UNIFORM,-10.0f*i,0.0f,-2000.0f);
//		renderObject.Shaders[WLF_SHADER]->assignUniform4f(WLF_GRATING_PARAM_UNIFORM, 0.1 + 0.001f*i,0.02f,0.0f,0.0f);
		
	//Do multipass rendering using diffraction shader
		finalRender = computeDiffractionTexture(renderObject, fname, grating1);
		}

//The rest of the stuff is not even called
    exit(0);   

    
//Now setup the final camera for rendering on the window
    setupFinalCamera(&cam1);
    
    unsigned int frameCount = 0;
    unsigned int nextUpdate = SDL_GetTicks() + 1000;
    /* Main loop */
    quit = 0;
    while (!quit)
		{
		while (SDL_PollEvent(&sdlEv))
			{
			switch (sdlEv.type)
				{
				case SDL_QUIT:
					quit = 1;
					cerr<<"end signal\n";
					break;
				
				case SDL_KEYDOWN:
					keyInput(sdlEv);
				default:
					break;
				}		
			}
		
		//Render our final texture	
		render(finalRender, renderObject.Shaders[FINAL_DRAW_SHADER]);	
	
		frameCount++;
		if (SDL_GetTicks() >= nextUpdate)
			{
			fprintf(stderr, "FPS: %d\n", frameCount);
			frameCount = 0;
			nextUpdate = SDL_GetTicks() + 1000;
			}
		}
    
    return EXIT_SUCCESS;
    }

//This is used to render the scene onto an FBO (bucketing and all)
void setupSceneCamera(Camera* cam)
	{
	glEnable (GL_DEPTH_TEST);
	
	glShadeModel(GL_SMOOTH);
	glViewport(0, 0, SCREEN_SIZE, SCREEN_SIZE);
 
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

#ifdef CAM_TYPE_ZOOMIN
	gluPerspective(40, 1, 1, 200);
	//zoomin camera
	cam->init(	0,0,-20,
				0,0,100,
				0,1,0);	
#elif defined CAM_TYPE_ZOOMIN_LOTS
	gluPerspective(40, 1, 0.9, 200);
	//zoomin camera
	cam->init(	0,0,-4,
				0,0,100,
				0,1,0);	
#elif defined CAM_TYPE_ZOOMIN_LOTS_MORE	
	gluPerspective(40, 1, 0.1, 20);
	//zoomin camera
	cam->init(	0,0,-1,
				0,0,100,
				0,1,0);	
#elif defined CAM_TYPE_ZOOMOUT
	gluPerspective(40, 1, 10, 2000);
	//zoomed out camera
	cam->init(	0,0,-100,
				0,0,100,
				0,1,0);		
#elif defined CAM_TYPE_ZOOMOUT_LOTS
	gluPerspective(40, 1, 10, 2000);
	//zoomed out camera
	cam->init(	0,0,-400,
				0,0,100,
				0,1,0);		
#elif defined CAM_TYPE_QUADRANT
	gluPerspective(40, 1, 10, 2000);
	//quadrant camera				
	cam->init(	50,50,-130,
				50,50,100,
				0,1,0);		
#elif defined CAM_TYPE_SEE_GRATING
	gluPerspective(20, 1, .1, 1000);
	//quadrant camera	
	cam->init(	-50,-50,-100,
				0,0,0,
				0,1,0);						
#endif				
	}

int initMain()
	{
//FIXME: cleanup previously initialised stuff after any error.
	cout<<"Initializing SDL"<<endl;
	bool sdlStat=sdlInit();
	if (sdlStat == true)
		{
	    cerr<<"SDL error\n";
	    return -1;
		}

	cout<<"Initializing glew"<<endl;
	glewInit();
	if (GLEW_VERSION_2_0)
		cerr<<"INFO: OpenGL 2.0 supported, proceeding"<<endl;
	else
		{
		cerr<<"INFO: OpenGL 2.0 not supported. Exit"<<endl;
		return -1;
		}			
	return 0;
	}



//This is used to draw the result finally onto the screen
void setupFinalCamera(Camera* cam)
	{
	glEnable (GL_DEPTH_TEST);
	
	glShadeModel(GL_SMOOTH);
	glViewport(0, 0, sdlWidth, sdlHeight);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(40, 1, 1, 1000);
	glMatrixMode(GL_MODELVIEW);

	//init camera
	cam->init(	128,128,-360,
				128,128,0,
				0,1,0);
	}

void keyInput(SDL_Event sdlEv)
	{
	switch (sdlEv.key.keysym.sym)
		{			
		case SDLK_ESCAPE:
			quit=1;
			break;
		case SDLK_UP:
			cam1.rotateCamera(RIGHT,0.1);
			break;						
		case SDLK_DOWN:
			cam1.rotateCamera(RIGHT,-0.1);
			break;						
		case SDLK_LEFT:
			cam1.rotateCamera(UP,0.1);
			break;						
		case SDLK_RIGHT:
			cam1.rotateCamera(UP,-0.1);
			break;		
		case SDLK_PAGEUP:
			cam1.moveCamera(UP,5);
			break;						
		case SDLK_PAGEDOWN:
			cam1.moveCamera(UP,-5);
			break;
		case SDLK_w:
			cam1.moveCamera(FORWARD,5);
			break;
		case SDLK_s:
			cam1.moveCamera(FORWARD,-5);
			break;
		case SDLK_a:
			cam1.moveCamera(RIGHT,-5);
			break;
		case SDLK_d:
			cam1.moveCamera(RIGHT,5);
			break;
		default:
			break;
		}
	}

void render(Texture* tex, ProgramGLSL* shader)
	{
//This renders the scene with specified texture, and shader
    glClearColor(0.5,0.5,0.5, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

//bind the shader       
    shader->bind();

//set the shader uniforms
	shader->setAllUniforms();	//use texture from unit 0
	
//bind textures
    glActiveTexture(GL_TEXTURE0);//set active texture unit as 0    
	tex->bind();//set the texture 

//Texture map a quad to show result
	drawQuad(256);
    
    tex->unbind();
    shader->unbind();
    
    SDL_GL_SwapBuffers();
    }


