/*
 * Copyright (C) 2008 Rivo Laks <rivolaks@hot.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either 
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include "program.h"

#include <shader.h>
#include <GL/glew.h>
#include <string>
#include<iostream>
using namespace std;

ProgramGLSL::ProgramGLSL()
{
    init();
}
/*
Program::Program(const QList<Shader*>& shaders)
{
    init();
    addShaders(shaders);
    link();
}
*/
ProgramGLSL::ProgramGLSL(const char * vertexshaderfile, const char * fragmentshaderfile)
{
    init();
    // Load both shaders, even if one is invalid. This way developer can see
    //  any errors/warnings for both shaders at once.
    VertexShader vs(vertexshaderfile);
    FragmentShader fs(fragmentshaderfile);
    if (!vs.isValid() || !fs.isValid()) {
        return;
    }
    addShader(&vs);
    addShader(&fs);
//    link();
}

ProgramGLSL::~ProgramGLSL()
{
    glDeleteProgram(glId());
//    cerr<<"ProgramGLSL: exiting, id\n"<<mGLId;
    delete[] mLinkLog;
    //delete mUniformLocations;
//    delete mAttributeLocations;
}

void ProgramGLSL::init()
{
    // Create program object
    mGLId = glCreateProgram();
    mValid = false;
    mLinkLog = 0;
//    mUniformLocations = mAttributeLocations = 0;
}

void ProgramGLSL::addShader(Shader* shader)
{
    if (!shader->isValid()) {
        cerr << "Program::addShader(): Cannot add invalid shader";
        return;
    }
    // Attach the shader to the program
    glAttachShader(glId(), shader->glId());
}
/*
void Program::addShaders(const QList<Shader*>& shaders)
{
    foreach (Shader* s, shaders) {
        addShader(s);
    }
}
*/
bool ProgramGLSL::link()
{
    // Link the program
    glLinkProgram(glId());
    // Make sure it linked correctly
    int linked;
    glGetProgramiv(glId(), GL_LINK_STATUS, &linked);
    mValid = linked;
    // Get info log
    GLsizei logsize, logarraysize;
    glGetProgramiv(glId(), GL_INFO_LOG_LENGTH, &logarraysize);
    mLinkLog = new char[logarraysize];
    glGetProgramInfoLog(glId(), logarraysize, &logsize, mLinkLog);
    if (!mValid) {
        cerr << "Program::link(): Couldn't link program. Log follows:" << endl << mLinkLog;
    } else {
//        mUniformLocations = new QHash<QString, int>;
//        mAttributeLocations = new QHash<QString, int>;
    }
    if (!logsize) {
        delete[] mLinkLog;
        mLinkLog = 0;
    }
    return mValid;
}

void ProgramGLSL::bind() const
{
	glUseProgram(this->glId());	
 //   renderer->bindProgram(this);
}

void ProgramGLSL::unbind() const
{
	glUseProgram(0);	
}

int ProgramGLSL::uniformLocation(const string& name)
{
    return glGetUniformLocation(glId(), name.c_str());
}

//Additions by kashyap
GLint ProgramGLSL::initUniform(const char * uniform_name, int uniform_id, int varCount, bool integerType)
{
	if (uniform_id >= (int)uniform_data.size())//allocate space
		uniform_data.resize(uniform_id+1);
	
	uniform_data[uniform_id].location = uniformLocation(uniform_name);
	uniform_data[uniform_id].varCount = varCount;
	uniform_data[uniform_id].integerType = integerType;
	return uniform_data[uniform_id].location;
}

void ProgramGLSL::assignUniform1i(int uniform_id, int value)
{
	uniform_data[uniform_id].iValue = value;
	uniform_data[uniform_id].dirty = true;
}
	
void ProgramGLSL::assignUniform1f(int uniform_id, float value)
{
	uniform_data[uniform_id].fValue[0]=value;
	uniform_data[uniform_id].dirty = true;
}

void ProgramGLSL::assignUniform2f(int uniform_id, float value1, float value2)
{
	uniform_data[uniform_id].fValue[0]=value1;
	uniform_data[uniform_id].fValue[1]=value2;		
	uniform_data[uniform_id].dirty = true;
}

void ProgramGLSL::assignUniform3f(const int uniform_id, const float value1, const float value2, const float value3)
{
	uniform_data[uniform_id].fValue[0]=value1;
	uniform_data[uniform_id].fValue[1]=value2;
	uniform_data[uniform_id].fValue[2]=value3;
	uniform_data[uniform_id].dirty = true;
}

void ProgramGLSL::assignUniform4f(int uniform_id, float value1, float value2, float value3, float value4)
{
	uniform_data[uniform_id].fValue[0]=value1;
	uniform_data[uniform_id].fValue[1]=value2;		
	uniform_data[uniform_id].fValue[2]=value3;
	uniform_data[uniform_id].fValue[3]=value4;		
	uniform_data[uniform_id].dirty = true;
}	

void ProgramGLSL::setAllUniforms(bool bindNow)
{//Note that you should bind the shader before setting params
	
	if (bindNow)
		bind();
	
	for (int i=0; i< (int)uniform_data.size(); i++)
	{
		GLint tmpLoc = uniform_data[i].location;
		if (tmpLoc == -1 || !uniform_data[i].dirty)//invalid address
			continue;
			
		if (uniform_data[i].integerType)
			glUniform1i(tmpLoc, uniform_data[i].iValue);
		else
		{
			switch (uniform_data[i].varCount)
			{
			case 4:
				glUniform4f(tmpLoc, uniform_data[i].fValue[0], uniform_data[i].fValue[1],
									uniform_data[i].fValue[2],	uniform_data[i].fValue[3]);
				break;
			case 3:
				glUniform3f(tmpLoc, uniform_data[i].fValue[0], uniform_data[i].fValue[1],
									uniform_data[i].fValue[2]);
				break;					
			case 2:
				glUniform2f(tmpLoc, uniform_data[i].fValue[0], uniform_data[i].fValue[1]);
				break;					
			case 1:
				glUniform1f(tmpLoc, uniform_data[i].fValue[0]);
				break;
			}
		}
		uniform_data[i].dirty = false;//this uniform is already set. dont set it again till it is modified
	}
	
	if (bindNow)
		unbind();
}


/*
int Program::uniformLocation(const char* name)
{
    if (!mValid) {
        return -1;
    }
    if (!mUniformLocations->contains(name)) {
        int location = glGetUniformLocation(glId(), name);
        mUniformLocations->insert(name, location);
    }
    return mUniformLocations->value(name);
}
int Program::attributeLocation(const QString& name)
{
    return attributeLocation(name.toLatin1().data());
}

int Program::attributeLocation(const char* name)
{
    if (!mValid) {
        return -1;
    }
    if (!mAttributeLocations->contains(name)) {
        int location = glGetAttribLocation(glId(), name);
        mAttributeLocations->insert(name, location);
    }
    return mAttributeLocations->value(name);
}
*/
/*
void Program::invalidateLocations()
{
    mUniformLocations->clear();
    mAttributeLocations->clear();
}

bool Program::setUniform(const char* name, float value)
{
    int location = uniformLocation(name);
    if (location >= 0) {
        glUniform1f(location, value);
    }
    return (location >= 0);
}

bool Program::setUniform(const char* name, Eigen::Vector2f value)
{
    int location = uniformLocation(name);
    if (location >= 0) {
        glUniform2fv(location, 1, value.data());
    }
    return (location >= 0);
}

bool Program::setUniform(const char* name, Eigen::Vector3f value)
{
    int location = uniformLocation(name);
    if (location >= 0) {
        glUniform3fv(location, 1, value.data());
    }
    return (location >= 0);
}

bool Program::setUniform(const char* name, Eigen::Vector4f value)
{
    int location = uniformLocation(name);
    if (location >= 0) {
        glUniform4fv(location, 1, value.data());
    }
    return (location >= 0);
}

bool Program::setUniform(const char* name, int value)
{
    int location = uniformLocation(name);
    if (location >= 0) {
        glUniform1i(location, value);
    }
    return (location >= 0);
}
*/
