/*
 * Copyright (C) 2008 Rivo Laks <rivolaks@hot.ee>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either 
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "texture.h"
#include<iostream>
using namespace std;

GLint determineFormat(GLint internalformat)
{
    switch (internalformat) {
        // From OpenGL 2.1 spec
        case GL_ALPHA:
        case GL_ALPHA4:
        case GL_ALPHA8:
        case GL_ALPHA12:
        case GL_ALPHA16:
            return GL_ALPHA;
        case GL_DEPTH_COMPONENT16:
        case GL_DEPTH_COMPONENT24:
        case GL_DEPTH_COMPONENT32:
            return GL_DEPTH_COMPONENT;
        case GL_LUMINANCE:
        case GL_LUMINANCE4:
        case GL_LUMINANCE8:
        case GL_LUMINANCE12:
        case GL_LUMINANCE16:
            return GL_LUMINANCE;
        case GL_LUMINANCE_ALPHA:
        case GL_LUMINANCE4_ALPHA4:
        case GL_LUMINANCE6_ALPHA2:
        case GL_LUMINANCE8_ALPHA8:
        case GL_LUMINANCE12_ALPHA4:
        case GL_LUMINANCE12_ALPHA12:
        case GL_LUMINANCE16_ALPHA16:
            return GL_LUMINANCE_ALPHA;
        case GL_INTENSITY:
        case GL_INTENSITY4:
        case GL_INTENSITY8:
        case GL_INTENSITY12:
        case GL_INTENSITY16:
            return GL_INTENSITY;
        case GL_R3_G3_B2:
        case GL_RGB:
        case GL_RGB4:
        case GL_RGB5:
        case GL_RGB8:
        case GL_RGB10:
        case GL_RGB12:
        case GL_RGB16:
            return GL_RGB;
        case GL_RGBA:
        case GL_RGBA2:
        case GL_RGBA4:
        case GL_RGB5_A1:
        case GL_RGBA8:
        case GL_RGB10_A2:
        case GL_RGBA12:
        case GL_RGBA16:
            return GL_RGBA;
        // sRGB
        case GL_SRGB8:              return GL_RGB;
        case GL_SRGB8_ALPHA8:       return GL_RGBA;
        case GL_SLUMINANCE:         return GL_LUMINANCE;
        case GL_SLUMINANCE8_ALPHA8: return GL_LUMINANCE_ALPHA;
        // Generic compressed formats
        case GL_COMPRESSED_ALPHA:           return GL_ALPHA;
        case GL_COMPRESSED_LUMINANCE:       return GL_LUMINANCE;
        case GL_COMPRESSED_LUMINANCE_ALPHA: return GL_LUMINANCE_ALPHA;
        case GL_COMPRESSED_INTENSITY:       return GL_INTENSITY;
        case GL_COMPRESSED_RGB:             return GL_RGB;
        case GL_COMPRESSED_RGBA:            return GL_RGBA;
        case GL_COMPRESSED_SRGB:            return GL_RGB;
        case GL_COMPRESSED_SRGB_ALPHA:      return GL_RGBA;
        case GL_COMPRESSED_SLUMINANCE:      return GL_LUMINANCE;
        case GL_COMPRESSED_SLUMINANCE_ALPHA: return GL_LUMINANCE_ALPHA;
        // From EXT_texture_compression_s3tc extension
        case GL_COMPRESSED_RGB_S3TC_DXT1_EXT:
            return GL_RGB;
        case GL_COMPRESSED_RGBA_S3TC_DXT1_EXT:
        case GL_COMPRESSED_RGBA_S3TC_DXT3_EXT:
        case GL_COMPRESSED_RGBA_S3TC_DXT5_EXT:
            return GL_RGBA;
        // From ARB_texture_float extension
        case GL_RGBA32F_ARB:                return GL_RGBA;
        case GL_RGB32F_ARB:                 return GL_RGB;
        case GL_ALPHA32F_ARB:               return GL_ALPHA;
        case GL_INTENSITY32F_ARB:           return GL_INTENSITY;
        case GL_LUMINANCE32F_ARB:           return GL_LUMINANCE;
        case GL_LUMINANCE_ALPHA32F_ARB:     return GL_LUMINANCE_ALPHA;
        case GL_RGBA16F_ARB:                return GL_RGBA;
        case GL_RGB16F_ARB:                 return GL_RGB;
        case GL_ALPHA16F_ARB:               return GL_ALPHA;
        case GL_INTENSITY16F_ARB:           return GL_INTENSITY;
        case GL_LUMINANCE16F_ARB:           return GL_LUMINANCE;
        case GL_LUMINANCE_ALPHA16F_ARB:     return GL_LUMINANCE_ALPHA;

        default:
            return 0;
    }
}

GLint determineDataType(GLint internalformat)
{
    switch (internalformat) {
        case GL_RGBA32F_ARB:
        case GL_RGB32F_ARB:
        case GL_ALPHA32F_ARB:
        case GL_INTENSITY32F_ARB:
        case GL_LUMINANCE32F_ARB:
        case GL_LUMINANCE_ALPHA32F_ARB:
            return GL_FLOAT;
        case GL_RGBA16F_ARB:
        case GL_RGB16F_ARB:
        case GL_ALPHA16F_ARB:
        case GL_INTENSITY16F_ARB:
        case GL_LUMINANCE16F_ARB:
        case GL_LUMINANCE_ALPHA16F_ARB:
            return GL_HALF_FLOAT_ARB;

        default:
            return GL_UNSIGNED_BYTE;
    }
}


/*** TextureBase ***/
TextureBase::TextureBase()
{
    mValid = false;
    mGLId = 0;
}

TextureBase::~TextureBase()
{
    if (mGLId) {
        glDeleteTextures(1, &mGLId);
    }
}

void TextureBase::bind() const
{
    glBindTexture(this->glTarget(), this->glId());
}

void TextureBase::unbind() const
{
    glBindTexture(this->glTarget(), 0);
}

void TextureBase::enable(bool _bind) const
{
    if (_bind) {
        bind();
    }
    glEnable(this->glTarget());    
}

void TextureBase::disable(bool _unbind) const
{
    glDisable(this->glTarget());
    if (_unbind) {
        unbind();
    }
}

void TextureBase::setFilter(GLenum filter)
{
    GLenum magfilter = (filter == GL_NEAREST) ? GL_NEAREST : GL_LINEAR;
    glTexParameteri(glTarget(), GL_TEXTURE_MAG_FILTER, magfilter);
    glTexParameteri(glTarget(), GL_TEXTURE_MIN_FILTER, filter);
}

void TextureBase::setWrapMode(GLenum coordinate, GLenum mode)
{
    glTexParameteri(glTarget(), coordinate, mode);
}

bool Texture::readBack(GLvoid * pixels) const
    {
    this->bind();
    glGetTexImage( GL_TEXTURE_2D,
	0,
	mFormat,//GL_RGBA
	determineDataType(mInternalFormat),//GL_FLOAT
	pixels);
    this->unbind();
    return false;
    }

/*** Texture ***/
Texture::Texture(int width, int height, GLint internalformat, GLint format,const GLvoid *pixels)
{
    mValid = init(width, height, internalformat, format,pixels);
}

Texture::~Texture()
{
//cerr<<"TEXTURE:texture exiting\n, id"<<mGLId;
}

bool Texture::init(int width, int height, GLint internalformat, GLint format,const GLvoid *pixels)
{
    mWidth = width;
    mHeight = height;

    glGenTextures(1, &mGLId);
    bind();
    setFilter(GL_LINEAR);

    if (!format) {
        format = determineFormat(internalformat);
        if (!format) {
            cerr << "Texture::init(): Couldn't determine matching format for internalformat" << internalformat;
            return false;
        }
    }
    mInternalFormat = internalformat;
    mFormat = format;
    glTexImage2D( glTarget(), 0, mInternalFormat, mWidth, mHeight, 0,
        mFormat, determineDataType(mInternalFormat), pixels);
/*
	cout<<"mInternalFormat:"<<mInternalFormat<<", mFormat:"<<mFormat<<", datatype:"<<determineDataType(mInternalFormat)<<endl;
	cout<<"mInternalFormat:"<<GL_RGBA32F_ARB<<", mFormat:"<<GL_RGBA<<", datatype:"<<GL_FLOAT<<endl;	
*/
      return true;
}

void Texture::setWrapMode(GLenum mode)
{
    TextureBase::setWrapMode(GL_TEXTURE_WRAP_S, mode);
    TextureBase::setWrapMode(GL_TEXTURE_WRAP_T, mode);
}

