#include"vbo.h"
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

GLuint VertexBuffer::createVBO(const void* data, int dataSize, GLenum target, GLenum usage){
	GLuint id = 0;  // 0 is reserved, glGenBuffersARB() will return non-zero id if success

	glGenBuffers(1, &id);                        // create a vbo
	glBindBuffer(target, id);                    // activate vbo id to use
	glBufferData(target, dataSize, data, usage); // upload data to video card

	// check data size in VBO is same as input array, if not return 0 and delete VBO
	int bufferSize = 0;
	glGetBufferParameteriv(target, GL_BUFFER_SIZE, &bufferSize);
	if(dataSize != bufferSize)
	{
	    glDeleteBuffers(1, &id);
	    id = 0;
	    cerr << "[createVBO()] Data size is mismatch with input array\n";
	}

	return id;      // return VBO id
}

VertexBuffer::VertexBuffer(){
	for (int i=0;i<6;i++){
		size[i]=offset[i]=0;
		flags[i]=false;
	}
	
	flags[F_INDEX]=true;//indexed by default
}

void VertexBuffer::enableEntity(eflags entity, int size, int offset){
	this->flags[entity]=true;
	this->size[entity]=size;
	this->offset[entity]=offset;
}

//creates cache friendly interleaved vbo
void VertexBuffer::init(const void* vertices,GLuint * indices,int vertexSize, int numVertices, int numIndices, GLenum mode){

    this->vertexSize=vertexSize;
    numberOfVertices=numVertices;
    int sizeVertices=vertexSize*numberOfVertices;
            
    int sizeIndices=numIndices*sizeof(GLuint);
    cout<<"Vertex Size: "<<vertexSize<<endl;
    numberOfIndices=numIndices;
    
    glGenBuffers(1, &vboId1);
    glBindBuffer(GL_ARRAY_BUFFER, vboId1);
    glBufferData(GL_ARRAY_BUFFER, sizeVertices, 0, GL_STATIC_DRAW);
  
    //all the vertex data is copied at once
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeVertices, vertices);
    
    // create VBO for index array
    // Target of this VBO is GL_ELEMENT_ARRAY_BUFFER_ARB and usage is GL_STATIC_DRAW_ARB
    vboId2 = createVBO(indices, sizeIndices, GL_ELEMENT_ARRAY_BUFFER, GL_STATIC_DRAW);
    
    dataFormat=mode;
}

void VertexBuffer::bind(){
    // bind VBOs with IDs and set the buffer offsets of the bound VBOs
    // When buffer object is bound with its ID, all pointers in gl*Pointer()
    // are treated as offset instead of real pointer.
    glBindBuffer(GL_ARRAY_BUFFER, vboId1);

	if (flags[F_NORMAL]){//Does not need size, normals always have 3 components
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT, vertexSize, (const GLvoid*)offset[F_NORMAL]);			
	}
	if (flags[F_COLOR]){
        glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(size[F_COLOR], GL_FLOAT, vertexSize, (const GLvoid*)offset[F_COLOR]);
	}
    if (flags[F_EDGE]){//Does not need size/type
        glEnableClientState(GL_EDGE_FLAG_ARRAY);
		glEdgeFlagPointer(vertexSize, (const GLvoid*)offset[F_EDGE]);	        
    }
    if (flags[F_INDEX]){//Enabled by default.
        glEnableClientState(GL_INDEX_ARRAY);
	}
	if (flags[F_VERTEX]){
        glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(size[F_VERTEX], GL_FLOAT, vertexSize, (const GLvoid*)offset[F_VERTEX]);
	}
	if (flags[F_TEXTURE]){
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(size[F_TEXTURE], GL_FLOAT, vertexSize, (const GLvoid*)offset[F_TEXTURE]);
	}

    // before draw, specify vertex and index arrays with their offsets
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboId2);
    glIndexPointer(GL_UNSIGNED_SHORT, 0, 0);
    //cout<<"Buffer bound"<<endl;

}

void VertexBuffer::render(bool bind_now){
	//cout<<"Drawing..."<<endl;
	if (bind_now)
		bind();
	
	glDrawElements(dataFormat,numberOfIndices,GL_UNSIGNED_INT,0);
	//cout<<"Drawn..."<<endl;
	
	if (bind_now)
		unbind();
}

void VertexBuffer::renderInstances(int numInstances, GLint instanceIdLoc=0){
//The function receives the location to uniform variable (glInstanceID) and updates it before each call.
//If instanceIdLoc is Zero, then it assumes hardware support exists for instancing.

	//Emulation Mode
	if (instanceIdLoc>0){
		for (int i=0;i<numInstances;i++){
			glUniform1i(instanceIdLoc,i);
			glDrawElements(dataFormat,numberOfIndices,GL_UNSIGNED_INT,0);
		}
		glUniform1i(instanceIdLoc,0);//reset instanceId to zero (this is there for compliance with the actual function)

	}else{	//Hardware Mode
	    glDrawElementsInstancedEXT(dataFormat, numberOfIndices, GL_UNSIGNED_INT, 0, numInstances);
	}
}

void VertexBuffer::unbind()
    {
    if  (flags[F_COLOR])	glDisableClientState(GL_COLOR_ARRAY);
    if  (flags[F_NORMAL]) 	glDisableClientState(GL_NORMAL_ARRAY);
    if  (flags[F_EDGE]) 	glDisableClientState(GL_EDGE_FLAG_ARRAY);               
    if  (flags[F_INDEX]) 	glDisableClientState(GL_INDEX_ARRAY);
    if  (flags[F_VERTEX]) 	glDisableClientState(GL_VERTEX_ARRAY);
    if  (flags[F_TEXTURE]) 	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    // it is good idea to release VBOs with ID 0 after use.
    // Once bound with 0, all pointers in gl*Pointer() behave as real
    // pointer, so, normal vertex array operations are re-activated
    glBindBuffer(GL_ARRAY_BUFFER_ARB, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
    }

///////////////////////////////////////////////////////////////////////////////
// destroy a VBO
// If VBO id is not valid or zero, then OpenGL ignores it silently.
///////////////////////////////////////////////////////////////////////////////
void VertexBuffer::deleteVBO(){
	glDeleteBuffers(1, &vboId1);
	glDeleteBuffers(1, &vboId2);
}

VertexBuffer::~VertexBuffer(){
	deleteVBO();
}

